@@ -61,6 +61,7 @@ function define_jfm(t)
    elseif type(t.zw)~='number' or type(t.zh)~='number' then
       defjfm_res= nil; return
    end
+   t.version = (type(t.version)=='number') and t.version or 1
    t.char_type = {}; t.chars = {}
    for i,v in pairs(t) do
       if type(i) == 'number' then -- char_type
@@ -86,7 +87,7 @@ function define_jfm(t)
 	    v.align = 'left' -- left
 	 end
 	 if type(v.width)~='number' then
-	    v.width = (jfm_dir=='tate') and  1.0
+	    v.width = nil
 	 end
 	 if type(v.height)~='number' then
 	    v.height = (jfm_dir=='tate') and  0.0
@@ -103,11 +104,49 @@ function define_jfm(t)
 	 if type(v.down)~='number' then
 	    v.down = 0.0
 	 end
+	 if t.version>=2 then
+	    if v.end_stretch then defjfm_res= nil; return end
+	    if v.end_shrink  then defjfm_res= nil; return end
+	    if v.end_adjust then
+	       if type(v.end_adjust)~='table' then
+	          v.end_adjust = nil
+	       elseif #(v.end_adjust)==0 then
+		  v.end_adjust = nil
+	       else 
+	          table.sort(v.end_adjust)
+	       end
+	    end
+	 else
+	    v.end_adjust = nil
+	    if v.end_stretch and v.end_stretch~=0.0 then 
+	       v.end_adjust = (v.end_adjust or {}) 
+	       v.end_adjust[#(v.end_adjust)+1] = v.end_stretch
+	    end
+	    if v.end_shrink and v.end_ahrink~=0.0 then 
+	       v.end_adjust = (v.end_adjust or {}) 
+	       v.end_adjust[#(v.end_adjust)+1] = -v.end_shrink
+	    end
+	    if v.end_adjust then v.end_adjust[#(v.end_adjust)+1] = 0.0 end
+	 end
 	 v.kern = v.kern or {}; v.glue = v.glue or {}
 	 for j,x in pairs(v.glue) do
 	    if v.kern[j] then defjfm_res= nil; return end
 	    x.ratio, x[5] = (x.ratio or (x[5] and 0.5*(1+x[5]) or 0.5)), nil
-	    x.priority, x[4] = (x.priority or x[4] or 0), nil
+	    do
+	       local xp
+	       xp, x[4] = (x.priority or x[4]), nil
+	       if type(xp)=='table' and t.version>=2 then
+		  if type(xp[1])~='number' or xp[1]<-4 or xp[1]>3 then defjfm_res=nil end  -- stretch
+		  if type(xp[2])~='number' or xp[2]<-4 or xp[2]>3 then defjfm_res=nil end  -- shrink
+		  xp = (xp[1]+4)*8+(xp[2]+4)
+	       elseif xp and type(xp)~='number' then
+		  defjfm_res = nil
+	       else
+		  xp = (xp or 0)*9+36	
+		  if xp<0 or xp>=64 then defjfm_res=nil end 
+	       end
+	       x.priority = xp
+	    end
 	    x.kanjiskip_natural = norm_val(x.kanjiskip_natural)
 	    x.kanjiskip_stretch = norm_val(x.kanjiskip_stretch)
 	    x.kanjiskip_shrink = norm_val(x.kanjiskip_shrink)
@@ -147,7 +186,6 @@ do
    end
    update_jfm_cache = function (j,sz)
       if metrics[j].size_cache[sz] then return end
-      --local TEMP = node_new(id_kern)
       local t = {}
       metrics[j].size_cache[sz] = t
       t.chars = metrics[j].chars
@@ -168,9 +206,8 @@ do
 	       }
 	    end
 	    for k,w in pairs(v.kern) do
-	       local g = node_new(id_kern)
+	       local g = node_new(id_kern, 1)
 	       setfield(g, 'kern', w[1])
-	       setfield(g, 'subtype', 1)
 	       set_attr(g, attr_icflag, FROM_JFM)
 	       v[k] = {g, ratio=w[2]/sz}
 	    end
@@ -182,7 +219,6 @@ do
       t.zw = round(metrics[j].zw*sz)
       t.zh = round(metrics[j].zh*sz)
       t.size = sz
-      --node_free(TEMP)
    end
 end
 
@@ -917,9 +953,9 @@ luatexbase.add_to_callback(
       local subtables = {}
       if ft.specification then
 	 for feat_name,v in pairs(ft.specification.features.normal) do
-	    if v==true then
+	    if v==true and ft.resources then
 	       for _,i in pairs(ft.resources.sequences) do
-		  if i.order[1]== feat_name and i.type == 'gpos_single' then
+		  if i.order[1]== feat_name and i.type == 'gpos_single' and type(i.subtables)=='table' then
 		     for _,st in pairs(i.subtables) do
 			subtables[st] = true
 		     end
@@ -949,8 +985,9 @@ do
   luatexbase.add_to_callback(
      'luaotfload.patch_font',
      function (tfmdata)
-	if tfmdata.cidinfo then
-	   local rd = ltjr_prepare_cid_font(tfmdata.cidinfo.registry, tfmdata.cidinfo.ordering)
+	local cidinfo = tfmdata.cidinfo or tfmdata.resources.cidinfo
+	if cidinfo and cidinfo.registry and cidinfo.ordering then
+	   local rd = ltjr_prepare_cid_font(cidinfo.registry, cidinfo.ordering)
 	   if rd then
 	      local ru, rc = rd.resources.unicodes, rd.characters
 	      for i,v in pairs(tfmdata.characters) do
@@ -985,8 +1022,8 @@ do
    else
        local ITALIC       = luatexja.icflag_table.ITALIC
        new_ic_kern = function()
-         local g = node_new(id_kern)
-         setfield(g, 'subtype', 1); set_attr(g, attr_icflag, ITALIC)
+         local g = node_new(id_kern, 1)
+         set_attr(g, attr_icflag, ITALIC)
 	 return g
        end
    end
