@@ -8,28 +8,25 @@ luatexja.load_module('stack');     local ltjs = luatexja.stack
 luatexja.load_module('jfont');     local ltjf = luatexja.jfont
 luatexja.load_module('direction'); local ltjd = luatexja.direction
 
-local Dnode = node.direct or node
+local to_node =  node.direct.tonode
+local to_direct =  node.direct.todirect
 
-local nullfunc = function(n) return n end
-local to_node = (Dnode ~= node) and Dnode.tonode or nullfunc
-local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
-
-local setfield = (Dnode ~= node) and Dnode.setfield or function(n, i, c) n[i] = c end
-local getid = (Dnode ~= node) and Dnode.getid or function(n) return n.id end
-local getfont = (Dnode ~= node) and Dnode.getfont or function(n) return n.font end
-local getchar = (Dnode ~= node) and Dnode.getchar or function(n) return n.char end
-local getfield = (Dnode ~= node) and Dnode.getfield or function(n, i) return n[i] end
-local getsubtype = (Dnode ~= node) and Dnode.getsubtype or function(n) return n.subtype end
+local setfield =  node.direct.setfield
+local getid =  node.direct.getid
+local getfont =  node.direct.getfont
+local getchar =  node.direct.getchar
+local getfield =  node.direct.getfield
+local getsubtype =  node.direct.getsubtype
 
 local pairs = pairs
 local floor = math.floor
-local has_attr = Dnode.has_attribute
-local set_attr = Dnode.set_attribute
-local node_traverse = Dnode.traverse
-local node_remove = Dnode.remove
-local node_next = (Dnode ~= node) and Dnode.getnext or node.next
-local node_free = Dnode.free
-local node_end_of_math = Dnode.end_of_math
+local has_attr = node.direct.has_attribute
+local set_attr = node.direct.set_attribute
+local node_traverse = node.direct.traverse
+local node_remove = node.direct.remove
+local node_next =  node.direct.getnext
+local node_free = node.direct.free
+local node_end_of_math = node.direct.end_of_math
 local tex_getcount = tex.getcount
 
 local id_glyph = node.id('glyph')
@@ -45,6 +42,7 @@ local attr_icflag = luatexbase.attributes['ltj@icflag']
 local is_ucs_in_japanese_char = ltjc.is_ucs_in_japanese_char_direct
 local ltjs_orig_char_table = ltjs.orig_char_table
 local ltjf_replace_altfont = ltjf.replace_altfont
+local ltjf_font_extra_info = ltjf.font_extra_info
 local attr_orig_char = luatexbase.attributes['ltj@origchar']
 local STCK  = luatexja.userid_table.STCK
 local DIR   = luatexja.userid_table.DIR
@@ -116,7 +114,7 @@ do
       return h
    end
 
-   luatexbase.add_to_callback('hyphenate',
+   ltjb.add_to_callback('hyphenate',
 			      function (head,tail)
 				 return suppress_hyphenate_ja(head)
 			      end,'ltj.hyphenate')
@@ -124,8 +122,8 @@ end
 
 -- mode: true iff this function is called from hpack_filter
 local ltjs_report_stack_level = ltjs.report_stack_level
-local ltjf_vert_form_table    = ltjf.vert_form_table
 local ltjf_font_metric_table  = ltjf.font_metric_table
+local font_getfont = font.getfont
 local function set_box_stack_level(head, mode)
    local box_set, cl = 0, tex.currentgrouplevel + 1
    for _,p  in pairs(wt) do
@@ -136,32 +134,31 @@ local function set_box_stack_level(head, mode)
       node_free(p)
    end
    if ltjs.list_dir == dir_tate then
-      for p in Dnode.traverse_id(id_glyph,to_direct(head)) do
+      for p in node.direct.traverse_id(id_glyph,to_direct(head)) do
          if (has_attr(p, attr_icflag) or 0)<=0 and getfield(p, 'lang')==lang_ja then
-            local pfn = has_attr(p, attr_curtfnt) or getfont(p)
             local pc = ltjs_orig_char_table[p]
-	    local nf = ltjf_replace_altfont(pfn, pc)
+	    local nf = ltjf_replace_altfont( has_attr(p, attr_curtfnt) or getfont(p) , pc)
 	    setfield(p, 'font', nf)
 	    if ltjf_font_metric_table[nf].vert_activated then
-	       pc = ltjf_vert_form_table [getchar(p)]
-	       if font.getfont(nf).characters[pc] then setfield(p, 'char', pc) end
+	       local pc = getchar(p)
+	       pc = (ltjf_font_extra_info[nf] and ltjf_font_extra_info[nf][pc] and ltjf_font_extra_info[nf][pc].vform)
+	       if pc and font_getfont(nf).characters[pc] then setfield(p, 'char', pc) end
 	    end
 	 end
       end
    end
-   --luatexja.ext_show_node_list(head, 'S> ', print)
    return head
 end
 
 -- CALLBACKS
-luatexbase.add_to_callback('hpack_filter',
+ltjb.add_to_callback('hpack_filter',
    function (head)
      return set_box_stack_level(head, true)
-   end,'ltj.hpack_filter_pre',1)
-luatexbase.add_to_callback('pre_linebreak_filter',
+   end,'ltj.set_stack_level',1)
+ltjb.add_to_callback('pre_linebreak_filter',
   function (head)
      return set_box_stack_level(head, false)
-  end,'ltj.pre_linebreak_filter_pre',1)
+  end,'ltj.set_stack_level',1)
 
 luatexja.pretreat = {
    set_box_stack_level = set_box_stack_level,
